"""Base objects for definding templates

Born from a [stackoverflow Q&A](https://stackoverflow.com/questions/69383734/templated
-object-generation-in-python/69384846)

>>> from embody.templater import Templater
>>> # the following template has templated dicts, strings, and lists
>>> template = {
...     'hello': '{name}',
...     'how are you': ['{verb}', 2, '{name} and {verb} again']
... }
>>> g = Templater.template_func(template=template)
>>> g(name='NAME', verb="VERB")
{'hello': 'NAME', 'how are you': ['VERB', 2, 'NAME and VERB again']}
>>> str(g.__signature__)
'(*, name, verb)'
"""

import string
from functools import partial
from inspect import Signature, Parameter
from operator import itemgetter
from typing import Callable, Any, TypeVar, Generator, Tuple, Dict, List, Iterable
from collections import namedtuple

T = TypeVar('T')
U = TypeVar('U')
K = TypeVar('K')
V = TypeVar('V')


def get_generator_return(gen: Generator[T, Any, U]) -> Tuple[Generator[T, Any, U], U]:
    return_value = None

    def inner():
        nonlocal return_value
        return_value = yield from gen

    gen_items = list(inner())

    def new_gen():
        yield from gen_items
        return return_value

    return new_gen(), return_value


def unique_list_conserving_order(iterable: Iterable) -> list:
    """List of unique elements in the original order
    >>> unique_list_conserving_order([3, 2, 3, 5, 1, 2, 6, 3, 5])
    [3, 2, 5, 1, 6]
    """
    seen = set()
    return [x for x in iterable if len(seen) < len(seen.add(x) or seen)]


# TemplateFunc: TypeAlias = Generator[str, None, Callable[..., T]]
TemplateFunc = Generator[str, None, Callable[..., T]]


class Templater:
    templater_registry: Dict[type, Callable[[Any], TemplateFunc]] = {}

    @classmethod
    def register(cls, handles_type: type):
        def decorator(f):
            cls.templater_registry[handles_type] = f
            return f

        return decorator

    @classmethod
    def template_func_generator(cls, template: T) -> TemplateFunc[T]:
        if type(template) in cls.templater_registry:
            template_factory = cls.templater_registry[type(template)]
            return template_factory(template)
        else:
            # an empty generator that returns a function that returns the template unchanged,
            # since we don't know how to handle it
            def just_return():
                return lambda: template
                yield  # this yield is needed to tell python that this is a generator

            return just_return()

    @classmethod
    def template_func(cls, template: T) -> Callable[..., T]:
        gen = cls.template_func_generator(template)
        params, f = get_generator_return(gen)

        parameters = unique_list_conserving_order(
            Parameter(name=param, kind=Parameter.KEYWORD_ONLY) for param in params
        )
        f.__signature__ = Signature(parameters)
        return f


@Templater.register(str)
def templated_string_func(template: str) -> TemplateFunc[str]:
    """A function making templated strings. Like template.format, but with a signature"""
    f = partial(str.format, template)
    yield from filter(None, map(itemgetter(1), string.Formatter().parse(template)))

    return f


@Templater.register(dict)
def templated_dict_func(template: Dict[K, V]) -> TemplateFunc[Dict[K, V]]:
    DictEntryInfo = namedtuple(
        'DictEntryInfo', ['key_func', 'value_func', 'key_args', 'value_args']
    )
    entries: List[DictEntryInfo] = []
    for key, value in template.items():
        key_params, key_template_func = get_generator_return(
            Templater.template_func_generator(key)
        )
        value_params, value_template_func = get_generator_return(
            Templater.template_func_generator(value)
        )
        key_params = tuple(key_params)
        value_params = tuple(value_params)
        yield from key_params
        yield from value_params

        entries.append(
            DictEntryInfo(
                key_template_func, value_template_func, key_params, value_params
            )
        )

    def template_func(**kwargs):
        return {
            entry_info.key_func(
                **{arg: kwargs[arg] for arg in entry_info.key_args}
            ): entry_info.value_func(
                **{arg: kwargs[arg] for arg in entry_info.value_args}
            )
            for entry_info in entries
        }

    return template_func


@Templater.register(list)
def templated_list_func(template: List[T]) -> TemplateFunc[List[T]]:
    entries = []
    for item in template:
        params, item_template_func = get_generator_return(
            Templater.template_func_generator(item)
        )
        params = tuple(params)
        yield from params

        entries.append((item_template_func, params))

    # How is this used?
    def template_func(**kwargs):
        # print(kwargs)
        return [
            item_template_func(**{arg: kwargs[arg] for arg in args})
            for item_template_func, args in entries
        ]

    return template_func



"""CA templating"""

from functools import partial
import operator as o


def example():
    # ------------------------The new syntax, essentially all you need to change on
    # your side------------------------------------------
    from verb import Command

    class PlaceHolder:
        pass

    ph = PlaceHolder()

    # This is how Andie defined the conditions:
    # phase_definitions = [{'phase_name': 'phase_1', 'tag_filters': [{'tag':
    # 'temperature', 'operator': '<', 'value': 80},
    #                                                                {'tag':
    #                                                                'temperature',
    #                                                                'operator': '>',
    #                                                                'value': 60}]},
    #                      {'phase_name': 'phase_2', 'tag_filters': [{'tag':
    #                      'temperature', 'operator': '>', 'value': 0}]}]

    # The same translated in Thor's language
    thored_phase_definitions = [
        {
            'phase_name': 'phase_1',
            'tag_filters': {
                '&': (
                    {'<': ({'[]': (ph, 'temperature')}, 80)},
                    {'>': ({'[]': (ph, 'temperature')}, 60)},
                )
            },
        },
        {
            'phase_name': 'phase_2',
            'tag_filters': {'<': ({'[]': (ph, 'temperature')}, 80)},
        },
    ]

    # ------------------------How to write and execute a command to test a phase
    # condition, basically copy that code------------------------------------------

    from verb import Literal

    # then say you want to test if a certain event satisfies the condition:
    event_to_test = {'temperature': 2}
    # for a somewhat technical reason, you need at the moment to add that Literal to
    # your event_to_test, Thor is/worked at removing that so it most likely will not
    # be needed by the time you get that, if you pull verb
    event_to_test = Literal(event_to_test)
    # I pick for the demo the conditions defining the first phase in the
    # thored_phase_definitions
    condition_to_test = thored_phase_definitions[0]['tag_filters']
    # we then can translate the conditions into a dictionary which Command can understand
    command_dict = recursive_format_iterable(
        condition_to_test,
        {
            dict: {
                'replace_dict': {PlaceHolder: event_to_test},
                'init_func': dict_init_func,
                'search_func': dict_search_func,
                'set_func': set_in_dict,
                'post_process_output': lambda x: x,
            },
            tuple: {
                'replace_dict': {PlaceHolder: event_to_test},
                'init_func': tuple_init_func,
                'search_func': tuple_search_func,
                'set_func': set_in_tuple,
                'post_process_output': lambda x: tuple(x),
            },
        },
    )
    # and finally, we can execute the command, returning whether or not event_to_test satisfies the conditions to belong to phase 1
    Command.from_dict(command_dict, func_of_key)()


func_of_key = {
    '&': o.and_,
    '>': o.gt,
    '<': o.lt,
    '>=': o.ge,
    '<=': o.le,
    '==': o.eq,
    '[]': lambda d, tag: d[tag],
}


def dict_search_func(item, iterable):
    return iterable[item]


def set_in_dict(iterable, item, val):
    iterable[item] = val


def dict_init_func(iterable):
    return dict()


class PlaceHolder_1:
    pass


class PlaceHolder_2:
    pass


ph_1 = PlaceHolder_1()
ph_2 = PlaceHolder_2()


def format_iterable(
    iterable,
    replace_dict={PlaceHolder_1: 'NEW', PlaceHolder_2: 'EVEN_NEWER'},
    init_func=dict_init_func,
    search_func=dict_search_func,
    set_func=set_in_dict,
    post_process_output=lambda x: x,
):
    """

    :param iterable:
    :param replace_dict:
    :param init_func:
    :param search_func:
    :param set_func:
    :param post_process_output:
    :return:

    Works by default for dicts
    >>> d = {'a': 1, 'b': (1,2), 'c': ph_1, 'd': ph_2}
    >>> format_iterable(d)
    {'a': 1, 'b': (1, 2), 'c': 'NEW', 'd': 'EVEN_NEWER'}

    >>> # And we can adapt to other iterables
    >>> def tuple_search_func(item, iterable):
    ...    return item

    >>> def set_in_tuple(iterable, item, val):
    ...    iterable.append(val)

    >>> def tuple_init_func(iterable):
    ...    return list()

    >>> t = (1, 'a', ph_1, 2, ph_2)
    >>> format_iterable(t,
    ...                 init_func=tuple_init_func,
    ...                 search_func=tuple_search_func,
    ...                 set_func=set_in_tuple,
    ...                 post_process_output=lambda x: tuple(x))
    (1, 'a', 'NEW', 2, 'EVEN_NEWER')

    """

    iterable_search_for_type = partial(search_func, iterable=iterable)
    new_iterable = init_func(iterable)

    for item in iterable:
        value = iterable_search_for_type(item)
        if type(value) in replace_dict:
            set_func(new_iterable, item, replace_dict[type(value)])
        else:
            set_func(new_iterable, item, value)

    return post_process_output(new_iterable)


def tuple_search_func(item, iterable):
    return item


def set_in_tuple(iterable, item, val):
    iterable.append(val)


def tuple_init_func(iterable):
    return list()


class TypeNotKnown(ValueError):
    pass


def format_multi_iterable(
    iterable,
    recipe={
        dict: {
            'replace_dict': {PlaceHolder_1: 'NEW', PlaceHolder_2: 'EVEN_NEWER'},
            'init_func': dict_init_func,
            'search_func': dict_search_func,
            'set_func': set_in_dict,
            'post_process_output': lambda x: x,
        },
        tuple: {
            'replace_dict': {
                PlaceHolder_1: 'NEW_TUPLE',
                PlaceHolder_2: 'EVEN_NEWER_TUPLE',
            },
            'init_func': tuple_init_func,
            'search_func': tuple_search_func,
            'set_func': set_in_tuple,
            'post_process_output': lambda x: tuple(x),
        },
    },
):
    """
    Same as format_iterable but can handle several types of iterables
    :param iterable:
    :param recipe:
    :return:

    >>> d = {'a': 1, 'b': (1,2), 'c': ph_1, 'd': ph_2}
    >>> format_multi_iterable(d)
    {'a': 1, 'b': (1, 2), 'c': 'NEW', 'd': 'EVEN_NEWER'}

    >>> t = (1, 'a', ph_1, 2, ph_2)
    >>> format_multi_iterable(t)
    (1, 'a', 'NEW_TUPLE', 2, 'EVEN_NEWER_TUPLE')

    """

    iterable_type = type(iterable)
    if iterable_type in recipe:
        return format_iterable(iterable, **recipe[iterable_type])
    else:
        raise TypeNotKnown('The type of your iterable is not in your recipe')


def recursive_format_iterable(
    iterable,
    recipe={
        dict: {
            'replace_dict': {PlaceHolder_1: 'NEW', PlaceHolder_2: 'EVEN_NEWER'},
            'init_func': dict_init_func,
            'search_func': dict_search_func,
            'set_func': set_in_dict,
            'post_process_output': lambda x: x,
        },
        tuple: {
            'replace_dict': {
                PlaceHolder_1: 'NEWTUPLE',
                PlaceHolder_2: 'EVEN_NEWER_TUPLE',
            },
            'init_func': tuple_init_func,
            'search_func': tuple_search_func,
            'set_func': set_in_tuple,
            'post_process_output': lambda x: tuple(x),
        },
    },
):
    iterable_type = type(iterable)
    if iterable_type in recipe:
        recipe_for_type = recipe[iterable_type]
        new_iterable = recipe_for_type['init_func'](iterable=iterable)
        default_set_func = recipe_for_type['set_func']
        default_search_func = partial(recipe_for_type['search_func'], iterable=iterable)
        default_post_process = recipe_for_type['post_process_output']
        default_replace_dict = recipe_for_type['replace_dict']
    else:
        raise TypeNotKnown('The type of your iterable is not in your recipe')

    for item in iterable:
        value = default_search_func(item)
        value_type = type(value)
        if value_type in recipe:
            new_value = recursive_format_iterable(value, recipe=recipe)
        elif value_type in default_replace_dict:
            new_value = default_replace_dict[value_type]
        else:
            new_value = value
        default_set_func(new_iterable, item, new_value)

    return default_post_process(new_iterable)



"""Making templates for various objects"""
import string
from itertools import chain
from functools import reduce


class Literal:
    def __init__(self, obj):
        self.obj = obj


class Field:
    def __init__(self, name: str):
        assert str.isidentifier(name), f'Needs to be a valid python identifier: {name}'
        self.name = name


string_formatter = string.Formatter()

dflt_gen_specs = (
    (
        lambda x: isinstance(x, str),
        dict(
            split=string_formatter.parse,
            mapper=lambda x: (x[0], Field(x[1]) if x[1] is not None else ''),
            redu=lambda iterable: ''.join(chain.from_iterable(iterable)),
        ),
    ),
    (lambda x: isinstance(x, list), dict(split=iter, redu=''.join)),
    (lambda x: isinstance(x, dict), dict(split=lambda x: x.items(), redu=''.join)),
)

NoSpecsFound = object()


def get_specs(template, gen_specs=None):
    gen_specs = gen_specs or dflt_gen_specs
    for cond, spec in gen_specs:
        if cond(template):
            yield spec
    return NoSpecsFound  # will be contained in .args of StopIteration error instance


def templated_gen(template, gen_specs=None):
    gen_specs = gen_specs or dflt_gen_specs
    specs = next(get_specs(template, gen_specs), NoSpecsFound)
    print(specs)
    if specs is not NoSpecsFound:
        split, mapper, redu = specs['split'], specs['mapper'], specs['redu']
        # return template, split, mapper, redu
        return redu(map(mapper, split(template)))
